/*
 * voshtomoc.c
 * pidgin plugin
 *
 * This a plugin for pidgin to run shell commands through an IM window
 *
 * Copyright (C) 2007 Ioan A. Sucan (http://ioan.sucan.info)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; This document is under the scope of
 * the version 2 of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see COPYING); if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "voshtomoc.h"

/**********************************************************/

static int        n_progs = 0;         /* number of available programs */
static program_t *progs   = NULL;      /* list of available programs   */


static user_t *users   = NULL;         /* the list of users */
static int     n_users = 0;            /* the number of users */

static PurpleCmdId reload_cmd;         /* registered command */

/**********************************************************/

#define DEBUG_USER_FUNCTION(uid)					\
    purple_debug(PURPLE_DEBUG_INFO, "voshtomoc", "%s for user %d (%s)\n", __func__, \
		 uid, uid >= 0 && uid < n_users ? users[uid].name : "NULL")

/* count how many programs a user could run */
static int count_user_programs(const char *user)
{
    int r = 0;    
    int i, j;
    for (i = 0 ; i < n_progs ; i++)
	if (progs[i].active)
	    for (j = 0; j < progs[i].n_allowed; j++)
		if (strcmp(progs[i].allowed[j], "*") == 0 ||
		    strcmp(user, progs[i].allowed[j]) == 0) ++r;
    return r;
}

static void configure_user_programs(const char *user, program_instance_t *user_prg)
{
    int n = 0;    
    int i, j;
    for (i = 0 ; i < n_progs ; i++)
	if (progs[i].active)
	    for (j = 0; j < progs[i].n_allowed; j++)
		if (strcmp(progs[i].allowed[j], "*") == 0 ||
		    strcmp(user, progs[i].allowed[j]) == 0) 
		{
		    purple_debug(PURPLE_DEBUG_INFO, "voshtomoc", "user '%s' can run '%s'\n", user, progs[i].name);
		    user_prg[n].rwpipe[0] = user_prg[n].rwpipe[1] = -1;
		    user_prg[n].pid = -1;
		    user_prg[n].running = 0;		    
		    user_prg[n].prog = i;
		    ++n;
		}
}


/* for every user, we keep an internal ID, to identify the program
 * this user is connected to */
static int get_user_id(const char *user, int add_if_not_found)
{    
    int i;
    purple_debug(PURPLE_DEBUG_INFO, "voshtomoc", "%s for '%s'\n", __func__, user);
    
    for (i = 0 ; i < n_users ; i++)
	if (strcmp(user, users[i].name) == 0) return i;

    /* add user if not found */
    if (add_if_not_found)
    {
	int np = count_user_programs(user);
	if (np > 0)
	{
	    int old_n_users = n_users;
	    n_users++;
	    users = (user_t*)realloc(users, n_users*sizeof(user_t));
	    users[old_n_users].name = strdup(user);
	    users[old_n_users].nprogs = np;	
	    users[old_n_users].prg = (program_instance_t*)malloc(users[old_n_users].nprogs * sizeof(program_instance_t));
	    configure_user_programs(user, users[old_n_users].prg);
	    users[old_n_users].index = 0;
	    return old_n_users;
	}
	else
	    purple_debug(PURPLE_DEBUG_INFO, "voshtomoc", "No programs allowed for '%s'\n", user);
    }
    
    return -1;
}

static void stop_user_programs(int id)
{    
    DEBUG_USER_FUNCTION(id);

    if (id < n_users && id >= 0)
    {
	int i;

	for (i = 0 ; i < users[id].nprogs ; ++i)
	    if (users[id].prg[i].running) 
	    {
		purple_debug(PURPLE_DEBUG_INFO, "voshtomoc", "Stopping program '%s' for user '%s'\n", progs[users[id].prg[i].prog].name, users[id].name);
		kill(users[id].prg[i].pid, SIGTERM);
		close(users[id].prg[i].rwpipe[0]);
		close(users[id].prg[i].rwpipe[1]);
		users[id].prg[i].running = 0;	    
	    }    
    }
}


/* remove the data for a specific user */
static void remove_user_id(int id)
{   
    DEBUG_USER_FUNCTION(id);

    /* optimistic, assume correct id and a succesful removal */
    if (id < n_users && id >= 0)
    {	
	stop_user_programs(id);
	
	n_users--;
	
	free(users[id].name);
	users[id].name = users[n_users].name;    
	
	users[id].index = users[n_users].index;
	free(users[id].prg);
	
	users[id].nprogs = users[n_users].nprogs;	
	users[id].prg = users[n_users].prg;

	users = (user_t*)realloc(users, n_users*sizeof(user_t));
    } 
}

/* check whether a specific buddy is allowed to execute some program
   with or without command syntax */
static int buddy_allow(int userid, int is_idle, int as_cmd)
{    
    if (userid >= 0 && userid < n_users)	
	if (users[userid].index >= 0 && users[userid].index < users[userid].nprogs)
	{
	    int p = users[userid].prg[users[userid].index].prog;
	    int res = (as_cmd ? progs[p].is_cmd : !progs[p].is_cmd) && (!progs[p].when_away || is_idle != 0);
	    purple_debug(PURPLE_DEBUG_INFO, "voshtomoc", "Checking if '%s' is allowed to send commands %s while we are %s to program '%s' : %s.\n",
			 users[userid].name, as_cmd ? "with prefix" : "as plain text", is_idle ? "idle" : "not idle", progs[p].name,
			 res ? "Yes" : "No");
	    return res;	    
	}
    return 0;
}


/*******************************************************/

/* retrieve the command from !CMD(<command>) 
 * caller must free the allocated memory */

static char* get_sent_command(char **buffer)
{
    char* cmd = NULL;
    
    if (buffer != NULL && *buffer != NULL)
    {
	int n = strlen(*buffer);
	int in_parse = 0;
	int in_retr = 0;
	int cmd_prf_len = strlen(CMD_PREFIX); 
	int parse_pos = -cmd_prf_len;	
	char wait_for = 0;
	int cmd_start = 0;
	int cmd_len = 0;
	int i;
	for (i = 0 ; i < n ; i++) {
	    if ((*buffer)[i] == CMD_PREFIX[0] && !in_parse && !in_retr)
	    {
		in_parse = 1;
		parse_pos = i;
		continue;
	    }
	    if (in_parse)
	    {
		if (i-parse_pos >= cmd_prf_len)
		{
		    int j;
		    wait_for = (*buffer)[i];
		    for (j = 0; j < CMD_SEP_COUNT; j++) 
			if ((*buffer)[i] == CMD_BEG[j])
			{
			    wait_for = CMD_END[j];
			    break;
			}
		    cmd_start = i + 1;
		    in_parse = 0;
		    in_retr = 1;
		    continue;
		} else
		    if ((*buffer)[i] != CMD_PREFIX[i-parse_pos]) in_parse = 0;
	    }
	    if (in_retr)
		if ((*buffer)[i] == wait_for)
		{
		    cmd_len = i - cmd_start;
		    break;
		}      
	}
	if (cmd_len > 0)
	{
	    cmd = (char*)malloc((cmd_len+1)*sizeof(char));
	    strncpy(cmd, &((*buffer)[cmd_start]), cmd_len);
	    cmd[cmd_len] = 0;
	}
    }
    return cmd;
}


/* Remove pidgin markers */

static char* strip_pidgin_tags(const char* buffer)
{
    int i;
    int j = 0;
    int ignore = 0;
    int n = strlen(buffer);
    char *result = (char*)malloc(sizeof(char)*(n + 1));
    for (i = 0 ; i < n ; i++)
    {
	if (buffer[i] == '<') ignore = 1; else
	    if (buffer[i] == '>') ignore = 0; else
		if (!ignore) result[j++] = buffer[i];
    }
    result[j] = 0;
    result = (char*)realloc(result, sizeof(char)*j);
    return result;
}


/* Find the process id for a user. If user was not seen before,
 * create the ID, run the application and set up the pipes */

static int get_id_and_setup_child_process(const char* sender)
{
    int id = get_user_id(sender, 1);
    if (id < 0) return -1;
    
    
    /* fork, exec & setup for the pipes */
    if (!users[id].prg[users[id].index].running)
    {
	int err;
	pid_t pid;
	int extra_pipe[2];
	int prg_id = users[id].prg[users[id].index].prog;
	purple_debug(PURPLE_DEBUG_INFO, "voshtomoc", "Running program '%s' for user '%s'\n", progs[prg_id].name, sender);
	err = pipe(users[id].prg[users[id].index].rwpipe);
	if (err == 0)
	{
	    err = pipe(extra_pipe);
	    if (err == 0)
	    {
		pid = fork();
		if (pid == -1)
		    err = 1;
	    }
	}
	if (err != 0)
	{
	    close(extra_pipe[0]); close(extra_pipe[1]);
	    close(users[id].prg[users[id].index].rwpipe[0]); close(users[id].prg[users[id].index].rwpipe[1]);
	    users[id].prg[users[id].index].rwpipe[0] = users[id].prg[users[id].index].rwpipe[1] = -1;
	} else 
	{
	    
	    if (pid == 0) // child
	    { 
		dup2(users[id].prg[users[id].index].rwpipe[1], 1);
		dup2(users[id].prg[users[id].index].rwpipe[1], 2);
		close(users[id].prg[users[id].index].rwpipe[1]);
		close(users[id].prg[users[id].index].rwpipe[0]);
		
		dup2(extra_pipe[0], 0);
		close(extra_pipe[0]);
		close(extra_pipe[1]);
		
		execlp(progs[prg_id].name, progs[prg_id].name, NULL);
		
		{	  
		    char err_str[512] = {0};
		    strerror_r(errno, err_str, sizeof(err_str));
		    fprintf(stderr, "execlp(%s) failed: %s\nPATH='%s'", 
			    progs[prg_id].name, err_str, getenv("PATH"));
		}
		
		_exit(errno);
		
	    }
	    else  // parent
	    {
		close(users[id].prg[users[id].index].rwpipe[1]);
		close(extra_pipe[0]);
		users[id].prg[users[id].index].rwpipe[1] = extra_pipe[1];
		users[id].prg[users[id].index].pid = pid;
		users[id].prg[users[id].index].running = 1;
		// we now write to pipes[id][1]  which is stdin to child
		// and read from pipes[id][0] which is stdout & stderr to child
	    }
	}
	
    }
    
    return id;
}

/* send a command to the pipe and get the reply */
static char* get_reply(int id, const char* command)
{
    char temp_buf[16384];
    if (command != NULL)
    {
	int err;
	float wait_time;
	struct timespec req;
	
	/* define the waiting time */
	wait_time = progs[users[id].prg[users[id].index].prog].wait_time;	
	req.tv_sec  = (int)floorf(wait_time);
	req.tv_nsec = (long)((double)(wait_time - floorf(wait_time)) * 1e9);
	
	purple_debug(PURPLE_DEBUG_INFO, "voshtomoc", "Executing command '%s' for user '%s' in program '%s' (waiting %d seconds and %d miliseconds)\n", 
		     command, users[id].name, progs[users[id].prg[users[id].index].prog].name, (int)req.tv_sec, (int)(req.tv_nsec / 1000000));
	
	/* construct the text to be sent to the external application */
	memset(temp_buf, 0, sizeof(temp_buf));
	if (progs[users[id].prg[users[id].index].prog].send_name)
	    snprintf(temp_buf, sizeof(temp_buf), "%s\n%s\n", users[id].name, command);
	else
	    snprintf(temp_buf, sizeof(temp_buf), "%s\n", command);

	/* send the text */
	err = write(users[id].prg[users[id].index].rwpipe[1], temp_buf, strlen(temp_buf));

	/* wait for a bit */
	nanosleep(&req, NULL);
	
	/* read the reply */
	memset(temp_buf, 0, sizeof(temp_buf));
	err = read(users[id].prg[users[id].index].rwpipe[0], temp_buf, sizeof(temp_buf));
	
	/* return the output data */
	return (strcmp(temp_buf, "<empty>\n") == 0) ? NULL : strdup(temp_buf);
    }
    return NULL;
}

/* load the plugin configuration file */
static void load_plugin_config(void)
{    
    char home[512] = {0};
    snprintf(home, sizeof(home), "%s/.purple/voshtomoc.conf", getenv("HOME"));
    FILE* file = fopen(home, "r");
    
    if (file != NULL)
    {
	char buffer[4096] = {0};
	int i, j;
	int r;
	
	r = fscanf(file, "%d", &n_progs);
	progs = (program_t*)malloc(n_progs*sizeof(program_t));
	for (i = 0; i < n_progs; i++)
	{
	    r = fscanf(file, "%s", buffer);
	    progs[i].active = buffer[0] != '#' ? 1 : 0;
	    progs[i].name = strdup(buffer + (1 - progs[i].active));
	    r = fscanf(file, "%d", &(progs[i].is_cmd));
	    r = fscanf(file, "%d", &(progs[i].send_name));
	    r = fscanf(file, "%d", &(progs[i].when_away));
	    r = fscanf(file, "%f", &(progs[i].wait_time));
	    r = fscanf(file, "%d", &(progs[i].n_allowed));
	    progs[i].allowed = (char**)malloc(progs[i].n_allowed*sizeof(char*));
	    for (j = 0 ; j < progs[i].n_allowed; j++)
	    {
		r = fscanf(file, "%s", buffer);
		progs[i].allowed[j] = strdup(buffer);
	    }
	}
	
	fclose(file);
	
	purple_debug(PURPLE_DEBUG_INFO, "voshtomoc", "Loaded configuration file\n");
    } else
	purple_debug(PURPLE_DEBUG_INFO, "voshtomoc", "No configuration file found\n");
}


/* save the plugin configuration */
static void save_plugin_config(void)
{

    char home[512] = {0};
    snprintf(home, sizeof(home), "%s/.purple/voshtomoc.conf", getenv("HOME"));
    FILE* file = fopen(home, "w");
    
    if (file != NULL)
    {
	int i, j;
	
	fprintf(file, "%d\n", n_progs);
	
	for (i = 0; i < n_progs; i++)
	{
	    if (progs[i].active)
		fprintf(file, "%s\n", progs[i].name);
	    else
		fprintf(file, "#%s\n", progs[i].name);
	    fprintf(file, "%d %d %d\n", progs[i].is_cmd, progs[i].send_name, progs[i].when_away);
	    fprintf(file, "%f\n", progs[i].wait_time);
	    fprintf(file, "%d\n", progs[i].n_allowed);
	    
	    for (j = 0 ; j < progs[i].n_allowed; j++)
		fprintf(file, "%s\n", progs[i].allowed[j]);
	}
	
	
	fclose(file);	
	purple_debug(PURPLE_DEBUG_INFO, "voshtomoc", "Saved configuration file\n");
    }
    else
	purple_debug(PURPLE_DEBUG_INFO, "voshtomoc", "Could not save configuration file\n");
    
}

/* free memory allocated for configuration */
static void free_plugin_config(void)
{
    purple_debug(PURPLE_DEBUG_INFO, "voshtomoc", "%s called\n", __func__);
    
    int i, j;
    for (i = 0 ; i < n_progs ; i++)
    {
	free(progs[i].name);
	for (j = 0 ; j < progs[i].n_allowed ; j++)
	    free(progs[i].allowed[j]);
	free(progs[i].allowed);
    }
    free(progs);
    n_progs = 0; progs = NULL;
}

/* close all running programs (children) */
static void close_plugin_pipes(void)
{
    purple_debug(PURPLE_DEBUG_INFO, "voshtomoc", "%s called\n", __func__);
    int i;
    for (i = 0; i < n_users; i++)
    { 
	stop_user_programs(i);
	free(users[i].name);
	free(users[i].prg);
    }
    free(users);
    n_users = 0; users = NULL;
}

/* free all allocated memory */
static void free_plugin_memory(void)
{
    purple_debug(PURPLE_DEBUG_INFO, "voshtomoc", "%s called\n", __func__);
    close_plugin_pipes();
    free_plugin_config();
}

static gboolean message_recv(PurpleAccount *account, char **sender, char **buffer, PurpleConversation *conv, PurpleMessageFlags *flags)
{
    char *reply = NULL;
    int is_idle = purple_presence_is_idle(purple_account_get_presence(account));
    
    if (*sender != NULL)
    {
	int skip_rest = 0;
	int user_id = get_user_id(*sender, 1);
	char *text = strip_pidgin_tags(*buffer);
	if (user_id >= 0 && text != NULL)
	{	    
	    /* process commands */
	    
	    purple_debug(PURPLE_DEBUG_INFO, "voshtomoc", "Received '%s'\n", text);
	    
	    if (strcmp(text, "!LIST") == 0)
	    {
		int i;
		reply = strdup("You are allowed to run:");
		for (i = 0 ; i < users[user_id].nprogs ; ++i)
		{
		    const int pos = strlen(reply);
		    const char *pnm = progs[users[user_id].prg[i].prog].name;
		    int r = users[user_id].prg[i].running;
		    const int mem = pos + strlen(pnm) + 3 + (r ? 9 : 0);
		    reply = (char*)realloc(reply, mem * sizeof(char));
		    if (r)
			sprintf(reply + pos, " %s(running)%c", pnm, i + 1 == users[user_id].nprogs ? '.' : ',');
		    else
			sprintf(reply + pos, " %s%c", pnm, i + 1 == users[user_id].nprogs ? '.' : ',');
		    reply[mem - 1] = 0;
		}
		const int pos = strlen(reply);
		const char *pnm = progs[users[user_id].prg[users[user_id].index].prog].name;
		const int mem = pos + strlen(pnm) + 21;
		reply = (char*)realloc(reply, mem * sizeof(char));
		sprintf(reply + pos, " Active program is %s.", pnm);
		reply[mem - 1] = 0;
		skip_rest = 1;
	    }

	    if (strcmp(text, "!NEXT") == 0)
	    {
		users[user_id].index = (users[user_id].index + 1) % users[user_id].nprogs;
		const char *pnm = progs[users[user_id].prg[users[user_id].index].prog].name;
		const int mem = strlen(pnm) + 20;
		reply = (char*)malloc(mem * sizeof(char));
		sprintf(reply, "Active program is %s.", pnm);
		reply[mem - 1] = 0;
		skip_rest = 1;		
	    }

	    if (strcmp(text, "!PREV") == 0)
	    {
		users[user_id].index--;
		if (users[user_id].index < 0) users[user_id].index = users[user_id].nprogs - 1;
		const char *pnm = progs[users[user_id].prg[users[user_id].index].prog].name;
		const int mem = strlen(pnm) + 20;
		reply = (char*)malloc(mem * sizeof(char));
		sprintf(reply, "Active program is %s.", pnm);
		reply[mem - 1] = 0;
		skip_rest = 1;
	    }

	    if (strcmp(text, "!STOP") == 0)
	    {
		stop_user_programs(user_id);
		reply = strdup("Stopped all programs");
		skip_rest = 1;		
	    }
	    
	    int attempts = 0;	    
	    while (!skip_rest && attempts < users[user_id].nprogs)
	    {		
		if (!skip_rest)	
		    if (buddy_allow(user_id, is_idle, 1))
		    {
			char *cmd = get_sent_command(&text);
			if (cmd != NULL)
			{
			    int id = get_id_and_setup_child_process(*sender);
			    reply = get_reply(id, cmd);
			    free(cmd);
			    skip_rest = 1;
			}
		    }
		
		if (!skip_rest)	
		    if (buddy_allow(user_id, is_idle, 0))
		    {
			int id = get_id_and_setup_child_process(*sender);
			reply = get_reply(id, text);
			skip_rest = 1;		    
		    }
		
		if (!skip_rest)
		{
		    users[user_id].index = (users[user_id].index + 1) % users[user_id].nprogs;
		    attempts++;		    
		}
	    }
	}
	
	
	if (text != NULL) free(text);
    }
    
    if (reply != NULL)
    {
	
	PurpleConversation *conv = 
	    purple_find_conversation_with_account(PURPLE_CONV_TYPE_IM,
						  *sender, account);
	
	if (conv == NULL)
	    conv = purple_conversation_new(PURPLE_CONV_TYPE_IM, account, *sender);
	PurpleConvIm *im = (conv != NULL) ? purple_conversation_get_im_data(conv) : NULL;
	if (im != NULL)
	{
	    int l = strlen(reply);
	    int s = 0;
	    while (l - s > MAX_MSG_SIZE)
	    {
		char backup = reply[s + MAX_MSG_SIZE];
		reply[s + MAX_MSG_SIZE] = 0;
		purple_conv_im_send(im, &(reply[s]));
		s = s + MAX_MSG_SIZE;
		reply[s] = backup;
	    }
	    if (l > s)
		purple_conv_im_send(im, &(reply[s]));
	}
	free(reply);
    }
    
    return FALSE;
}

static void buddy_sign_off(PurpleBuddy *buddy)
{
    int buddy_id = get_user_id((const char*)purple_buddy_get_name(buddy), 0);    
    if (buddy_id >= 0) remove_user_id(buddy_id);
}


/* handler for the \vreload command. We reload the plugin configuration */
static PurpleCmdRet reload_plugin_config(PurpleConversation* conv, const gchar* cmd,
					 gchar** args, gchar** err, void* data) {
    free_plugin_memory();
    load_plugin_config();
    
    return PURPLE_CMD_RET_OK;
}

static gboolean plugin_load(PurplePlugin *plugin)
{

    void *conv_handle = purple_conversations_get_handle();
    
    purple_signal_connect(conv_handle, "receiving-im-msg",
			  plugin, PURPLE_CALLBACK(message_recv), NULL);
    
    void *blist_handle = purple_blist_get_handle();
    
    purple_signal_connect(blist_handle, "buddy-signed-off",
			  plugin, PURPLE_CALLBACK(buddy_sign_off), NULL);

    reload_cmd = purple_cmd_register("vreload", "", PURPLE_CMD_P_DEFAULT, PURPLE_CMD_FLAG_IM,
				     NULL, reload_plugin_config, "vreload:  Reload the configuration file for voshtomoc.", NULL);
    
    load_plugin_config();

    purple_debug(PURPLE_DEBUG_INFO, "voshtomoc", "Plugin loaded\n");

    return TRUE;
}

static gboolean plugin_unload(PurplePlugin *plugin)
{
    void *conv_handle = purple_conversations_get_handle();
    purple_signal_disconnect(conv_handle, "receiving-im-msg", plugin, PURPLE_CALLBACK(message_recv));
    
    void *blist_handle = purple_blist_get_handle();
    purple_signal_disconnect(blist_handle, "buddy-signed-off", plugin, PURPLE_CALLBACK(buddy_sign_off));
    
    purple_cmd_unregister(reload_cmd);
    
    free_plugin_memory();
    
    purple_debug(PURPLE_DEBUG_INFO, "voshtomoc", "Plugin unloaded\n");
        
    return TRUE;
}

/*********************************************************************/
/*                        Graphical configuration  		     */
/* (code inspired from the text replacement plugin)                  */
/*********************************************************************/

enum
{
    COL_NAME = 0,
    COL_ACTIVE,
    COL_IS_CMD,
    COL_SEND_NAME,
    COL_WHEN_IDLE,
    COL_WAIT_TIME,
    COL_USERS,
    NUM_COLS
} ;

static GtkListStore *model_prg;
static GtkWidget    *tree_prg;
static GtkWidget    *name_entry;
static GtkWidget    *user_entry;

static guint string_to_uint(gchar *string)
{
    int i;
    int n = strlen(string);
    guint v = 0;    
    for (i = 0 ; i < n ; i++)
	v = v * 10 + string[i] - '0';	
    return v;
}

static char* prog_user_list(int prgid)
{
    int j;
    char *usr_list = strdup("");
    for (j = 0 ; j < progs[prgid].n_allowed ; ++j)
    {
	const int pos = strlen(usr_list);
	const char *pnm = progs[prgid].allowed[j];
	const int extra = (j + 1) != progs[prgid].n_allowed ? 2 : 0;
	const int mem = pos + strlen(pnm) + 1 + extra;
	usr_list = (char*)realloc(usr_list, mem * sizeof(char));
	if (extra)
	    sprintf(usr_list + pos, "%s, ", pnm);
	else
	    sprintf(usr_list + pos, "%s", pnm);
	usr_list[mem - 1] = 0;
    }
    return usr_list;
}

static void create_and_fill_model(void)
{
    int            i;
    GtkTreeIter    iter;
    
    model_prg = gtk_list_store_new ((gint)NUM_COLS, G_TYPE_STRING, G_TYPE_BOOLEAN,
				    G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN,
				    G_TYPE_STRING, G_TYPE_STRING);
    
    /* Append a row and fill in some data */
    for (i = 0 ; i < n_progs ; i++)
    {
	char *usr_list = prog_user_list(i);
	char wait_time[64];
	snprintf(wait_time, sizeof(wait_time), "%0.3f", progs[i].wait_time);
	
	gtk_list_store_append (model_prg, &iter);
	gtk_list_store_set (model_prg, &iter,
			    COL_NAME, progs[i].name,
			    COL_ACTIVE, progs[i].active ? TRUE : FALSE,
			    COL_IS_CMD, progs[i].is_cmd ? TRUE : FALSE,
			    COL_SEND_NAME, progs[i].send_name ? TRUE : FALSE,
			    COL_WHEN_IDLE, progs[i].when_away ? TRUE : FALSE,
			    COL_WAIT_TIME, wait_time,
			    COL_USERS, usr_list,
			    -1);
	free(usr_list);	
    }
}

static void on_change_program_name(GtkCellRendererText *cellrenderertext,
				   gchar *path, gchar *arg2, gpointer data)
{
    GtkTreeIter iter;
    GValue val;
    guint prgid;
    
    if (arg2[0] == '\0')
    {
	gdk_beep();
	return;
    }
    
    g_return_if_fail(gtk_tree_model_get_iter_from_string(GTK_TREE_MODEL(model_prg), &iter, path));
    
    val.g_type = 0;
    gtk_tree_model_get_value(GTK_TREE_MODEL(model_prg), &iter, 0, &val);
    if (strcmp(arg2, g_value_get_string(&val)))
    {
	gtk_list_store_set(model_prg, &iter, 0, arg2, -1);
	prgid = string_to_uint(path);
	close_plugin_pipes();
	free(progs[prgid].name);
	progs[prgid].name = strdup(arg2);
	save_plugin_config();	
    }
    g_value_unset(&val);
}

static void on_active_toggled(GtkCellRendererToggle *cellrenderertoggle,
			      gchar *path, gpointer data)
{
    GtkTreeIter iter;
    gboolean enabled;
    guint prgid;

    g_return_if_fail(gtk_tree_model_get_iter_from_string(GTK_TREE_MODEL(model_prg), &iter, path));
    
    gtk_tree_model_get(GTK_TREE_MODEL(model_prg), &iter,
		       COL_ACTIVE, &enabled,
		       -1);
    enabled = !enabled;    
    gtk_list_store_set(GTK_LIST_STORE(model_prg), &iter,
		       COL_ACTIVE, enabled,
		       -1);

    prgid = string_to_uint(path);

    close_plugin_pipes();
    progs[prgid].active = 1 - progs[prgid].active;
    save_plugin_config();    
}

static void on_iscmd_toggled(GtkCellRendererToggle *cellrenderertoggle,
			     gchar *path, gpointer data)
{   
    GtkTreeIter iter;
    gboolean enabled;
    guint prgid;

    g_return_if_fail(gtk_tree_model_get_iter_from_string(GTK_TREE_MODEL(model_prg), &iter, path));
    
    gtk_tree_model_get(GTK_TREE_MODEL(model_prg), &iter,
		       COL_IS_CMD, &enabled,
		       -1);
    enabled = !enabled;    
    gtk_list_store_set(GTK_LIST_STORE(model_prg), &iter,
		       COL_IS_CMD, enabled,
		       -1);

    prgid = string_to_uint(path);

    close_plugin_pipes();
    progs[prgid].is_cmd = 1 - progs[prgid].is_cmd;    
    save_plugin_config();
}

static void on_sendname_toggled(GtkCellRendererToggle *cellrenderertoggle,
				gchar *path, gpointer data)
{
    GtkTreeIter iter;
    gboolean enabled;
    guint prgid;

    g_return_if_fail(gtk_tree_model_get_iter_from_string(GTK_TREE_MODEL(model_prg), &iter, path));
    
    gtk_tree_model_get(GTK_TREE_MODEL(model_prg), &iter,
		       COL_SEND_NAME, &enabled,
		       -1);
    enabled = !enabled;    
    gtk_list_store_set(GTK_LIST_STORE(model_prg), &iter,
		       COL_SEND_NAME, enabled,
		       -1);

    prgid = string_to_uint(path);

    close_plugin_pipes();	
    progs[prgid].send_name = 1 - progs[prgid].send_name;
    save_plugin_config();
}

static void on_whenidle_toggled(GtkCellRendererToggle *cellrenderertoggle,
				gchar *path, gpointer data)
{   
    GtkTreeIter iter;
    gboolean enabled;
    guint prgid;

    g_return_if_fail(gtk_tree_model_get_iter_from_string(GTK_TREE_MODEL(model_prg), &iter, path));
    
    gtk_tree_model_get(GTK_TREE_MODEL(model_prg), &iter,
		       COL_WHEN_IDLE, &enabled,
		       -1);
    enabled = !enabled;    
    gtk_list_store_set(GTK_LIST_STORE(model_prg), &iter,
		       COL_WHEN_IDLE, enabled,
		       -1);

    prgid = string_to_uint(path);

    close_plugin_pipes();	
    progs[prgid].when_away = 1 - progs[prgid].when_away;
    save_plugin_config();
}

static void on_change_wait_time(GtkCellRendererText *cellrenderertext,
				gchar *path, gchar *arg2, gpointer data)
{  
    GtkTreeIter iter;
    GValue val;
    guint prgid;
    
    if (arg2[0] == '\0')
    {
	gdk_beep();
	return;
    }
    
    g_return_if_fail(gtk_tree_model_get_iter_from_string(GTK_TREE_MODEL(model_prg), &iter, path));
    
    val.g_type = 0;
    gtk_tree_model_get_value(GTK_TREE_MODEL(model_prg), &iter, 5, &val);
    if (strcmp(arg2, g_value_get_string(&val)))
    {
	gtk_list_store_set(model_prg, &iter, 5, arg2, -1);
	prgid = string_to_uint(path);
	progs[prgid].wait_time = atof(arg2);
	save_plugin_config();	
    }
    g_value_unset(&val);
}

static void build_allowed_list(program_t *p, const char *arg)
{
    int i, n;
    
    if (p->n_allowed > 0)
    {
	for (i = 0 ; i < p->n_allowed ; ++i)
	    free(p->allowed[i]);
	free(p->allowed);
    }    
    p->allowed = NULL;
    p->n_allowed = 0;

    n = strlen(arg);
	
    {
	const int blen = 4096;	    
	char buffer[blen];
	int l_usr = 0;
	
	for (i = 0 ; i < n ; ++i)
	{
	    if (arg[i] == '\t' || arg[i] == '\n' ||
		arg[i] == '\r' || arg[i] == ' ') continue;
	    if (arg[i] == ',' || arg[i] == ';' || (i + 1 == n))
	    {
		if (arg[i] != ',' && arg[i] != ';')
		{
		    buffer[l_usr] = arg[i];
		    if (l_usr < blen) l_usr++;
		}
		
		if (l_usr > 0)
		{
		    buffer[l_usr] = '\0';
		    p->n_allowed++;
		    
		    p->allowed = (char**)realloc(p->allowed, 
						 sizeof(char*) * p->n_allowed);
		    p->allowed[p->n_allowed - 1] = strdup(buffer);
		    l_usr = 0;
		}
		continue;		    
	    }
	    buffer[l_usr] = arg[i];
	    if (l_usr < blen) l_usr++;		
	}
    }
}


static void on_change_program_users(GtkCellRendererText *cellrenderertext,
				    gchar *path, gchar *arg2, gpointer data)
{
    GtkTreeIter iter;
    GValue val;
    guint prgid;
    
    if (arg2[0] == '\0')
    {
	gdk_beep();
	return;
    }
    
    g_return_if_fail(gtk_tree_model_get_iter_from_string(GTK_TREE_MODEL(model_prg), &iter, path));
    
    val.g_type = 0;
    gtk_tree_model_get_value(GTK_TREE_MODEL(model_prg), &iter, 6, &val);

    if (strcmp(arg2, g_value_get_string(&val)))
    {
	prgid = string_to_uint(path);
	close_plugin_pipes();
	build_allowed_list(progs + prgid, arg2);
	
	char *usr_list = prog_user_list(prgid);
	gtk_list_store_set(model_prg, &iter, 6, usr_list, -1);
	free(usr_list);
	
	save_plugin_config();
    }
    g_value_unset(&val);
}

static void program_add_new(void)
{
    program_t p;
    const char *prog_name = gtk_entry_get_text(GTK_ENTRY(name_entry));
    const char *user_name = gtk_entry_get_text(GTK_ENTRY(user_entry));
    GtkTreeIter iter;
    
    p.name = strdup(prog_name);
    p.active = 0;
    p.is_cmd = 1;
    p.send_name = 0;
    p.when_away = 0;
    p.wait_time = 0.1f;
    p.n_allowed = 0;
    build_allowed_list(&p, user_name);
    
    n_progs++;
    progs = (program_t*)realloc(progs, sizeof(program_t) * n_progs);
    progs[n_progs-1] = p;
    
    gtk_list_store_append (model_prg, &iter);
    gtk_list_store_set (model_prg, &iter,
			COL_NAME, p.name,
			COL_ACTIVE, p.active ? TRUE : FALSE,
			COL_IS_CMD, p.is_cmd ? TRUE : FALSE,
			COL_SEND_NAME, p.send_name ? TRUE : FALSE,
			COL_WHEN_IDLE, p.when_away ? TRUE : FALSE,
			COL_WAIT_TIME, "0.100",
			COL_USERS, user_name,
			-1);
    
    gtk_editable_delete_text(GTK_EDITABLE(name_entry), 0, -1);
    gtk_editable_delete_text(GTK_EDITABLE(user_entry), 0, -1);
    gtk_widget_grab_focus(name_entry);

    save_plugin_config();    
}

static gboolean non_empty(const char *s)
{
    while (*s && g_ascii_isspace(*s))
	s++;
    return *s;
}

static void on_entry_changed(GtkEditable *editable, gpointer data)
{
    gtk_widget_set_sensitive((GtkWidget*)data,
			     non_empty(gtk_entry_get_text(GTK_ENTRY(name_entry))) &&
			     non_empty(gtk_entry_get_text(GTK_ENTRY(user_entry))));
}

static void select_program(int progid)
{
    // selected a program; may want to add some code in the future.
}

static void select_program_f(GtkTreeModel *model, GtkTreePath *path,
			     GtkTreeIter *iter, gpointer user_data)
{    
    int progid;    
    gchar *spath = gtk_tree_path_to_string(path);
    progid = string_to_uint(spath);
    g_free(spath);
    select_program(progid);
}

static void on_selection_changed(GtkTreeSelection *sel,
				 gpointer data)
{
    gtk_tree_selection_selected_foreach(sel, select_program_f, NULL);
}


static void rem_program(int id)
{
    int i;
    close_plugin_pipes();

    if (id < n_progs && id >= 0)
    {
	n_progs--;
	free(progs[id].name);
	progs[id].name = progs[n_progs].name;	
	progs[id].is_cmd = progs[n_progs].is_cmd;
	progs[id].send_name = progs[n_progs].send_name;
	progs[id].when_away = progs[n_progs].when_away;
	progs[id].active = progs[n_progs].active;
	progs[id].wait_time = progs[n_progs].wait_time;
	for (i = 0 ; i < progs[id].n_allowed ; i++)
	    free(progs[id].allowed[i]);
	free(progs[id].allowed);
	progs[id].n_allowed = progs[n_progs].n_allowed;	
	progs[id].allowed = progs[n_progs].allowed;
	
	progs = (program_t*)realloc(progs, n_progs*sizeof(program_t));
	save_plugin_config();  
    }
}

static void rem_program_f(GtkTreeModel *model, GtkTreePath *path,
			  GtkTreeIter *iter, gpointer user_data)
{    
    int progid;    
    gchar *spath = gtk_tree_path_to_string(path);
    progid = string_to_uint(spath);
    g_free(spath);
    rem_program(progid);
    gtk_list_store_remove(model_prg, iter);
}

static void prog_delete(void)
{
    GtkTreeSelection *sel = gtk_tree_view_get_selection(GTK_TREE_VIEW(tree_prg));
    gtk_tree_selection_selected_foreach(sel, rem_program_f, NULL);
}

static GtkWidget* create_view_and_model (void)
{
    GtkCellRenderer     *renderer;
    GtkWidget           *view;
    
    create_and_fill_model();

    view = gtk_tree_view_new_with_model (GTK_TREE_MODEL(model_prg));    
    g_signal_connect(G_OBJECT(gtk_tree_view_get_selection(GTK_TREE_VIEW(view))),
		     "changed", G_CALLBACK(on_selection_changed), NULL);

    renderer = gtk_cell_renderer_text_new ();
    g_object_set(G_OBJECT(renderer), "editable", TRUE, NULL);
    g_signal_connect(G_OBJECT(renderer), "edited",
		     G_CALLBACK(on_change_program_name), NULL);
    gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (view),
						 -1,      
						 "Name",  
						 renderer,
						 "text", COL_NAME,
						 NULL);
    
    
    renderer = gtk_cell_renderer_toggle_new ();
    g_object_set(G_OBJECT(renderer), "activatable", TRUE, NULL);
    g_signal_connect(G_OBJECT(renderer), "toggled",
		     G_CALLBACK(on_active_toggled), NULL);
    gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (view),
						 -1,      
						 "Active",
						 renderer,
						 "active", COL_ACTIVE,
						 NULL);
    
    renderer = gtk_cell_renderer_toggle_new ();
    g_object_set(G_OBJECT(renderer), "activatable", TRUE, NULL);
    g_signal_connect(G_OBJECT(renderer), "toggled",
		     G_CALLBACK(on_iscmd_toggled), NULL);
    gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (view),
						 -1,      
						 "Only command",
						 renderer,
						 "active", COL_IS_CMD,
						 NULL);

    renderer = gtk_cell_renderer_toggle_new ();
    g_object_set(G_OBJECT(renderer), "activatable", TRUE, NULL);
    g_signal_connect(G_OBJECT(renderer), "toggled",
		     G_CALLBACK(on_sendname_toggled), NULL);
    gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (view),
						 -1,      
						 "Send name",
						 renderer,
						 "active", COL_SEND_NAME,
						 NULL);

    renderer = gtk_cell_renderer_toggle_new ();
    g_object_set(G_OBJECT(renderer), "activatable", TRUE, NULL);
    g_signal_connect(G_OBJECT(renderer), "toggled",
		     G_CALLBACK(on_whenidle_toggled), NULL);
    gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (view),
						 -1,      
						 "When idle",
						 renderer,
						 "active", COL_WHEN_IDLE,
						 NULL);   

    renderer = gtk_cell_renderer_text_new ();
    g_object_set(G_OBJECT(renderer), "editable", TRUE, NULL);
    g_signal_connect(G_OBJECT(renderer), "edited",
		     G_CALLBACK(on_change_wait_time), NULL);
    gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (view),
						 -1,      
						 "Wait time",  
						 renderer,
						 "text", COL_WAIT_TIME,
						 NULL);

    renderer = gtk_cell_renderer_text_new ();
    g_object_set(G_OBJECT(renderer), "editable", TRUE, NULL);
    g_signal_connect(G_OBJECT(renderer), "edited",
		     G_CALLBACK(on_change_program_users), NULL);
    gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (view),
						 -1,      
						 "Users",  
						 renderer,
						 "text", COL_USERS,
						 NULL);

    gtk_tree_selection_set_mode(gtk_tree_view_get_selection(GTK_TREE_VIEW(view)),
				GTK_SELECTION_BROWSE);

    return view;
}


static GtkWidget* get_config_frame(PurplePlugin *plugin)
{
    GtkWidget *win, *ret, *vbox;
    GtkWidget *button, *label, *hbox;    
    GtkSizeGroup *sg;
    GtkSizeGroup *sg2;
    GtkWidget *vbox2;
    GtkWidget *hbox2;
    GtkWidget *vbox3;
    
    ret = gtk_vbox_new(FALSE, PIDGIN_HIG_CAT_SPACE);
    gtk_container_set_border_width (GTK_CONTAINER(ret), PIDGIN_HIG_BORDER);
    
    vbox = pidgin_make_frame(ret, "Programs");
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
    gtk_widget_show(vbox);

    win = gtk_scrolled_window_new(0, 0);
    gtk_box_pack_start(GTK_BOX(vbox), win, TRUE, TRUE, 0);
    gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(win),
					GTK_SHADOW_IN);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(win),
				   GTK_POLICY_ALWAYS,
				   GTK_POLICY_ALWAYS);
    gtk_widget_show(win);
    
    tree_prg = create_view_and_model();
    gtk_widget_set_size_request(tree_prg, 500, 200);
    
    gtk_container_add(GTK_CONTAINER(win), tree_prg);
    gtk_widget_show(tree_prg);
    
    hbox = gtk_hbutton_box_new();
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);
    
    button = gtk_button_new_from_stock(GTK_STOCK_DELETE);
    g_signal_connect(G_OBJECT(button), "clicked",
		     G_CALLBACK(prog_delete), NULL);
    gtk_box_pack_start(GTK_BOX(hbox), button, FALSE, FALSE, 0);
    
    gtk_widget_show(button);    
    
    vbox = pidgin_make_frame(ret, "Add a new program");

    hbox = gtk_hbox_new(FALSE, PIDGIN_HIG_BOX_SPACE);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);
    gtk_widget_show(hbox);
    vbox2 = gtk_vbox_new(FALSE, PIDGIN_HIG_BOX_SPACE);
    gtk_box_pack_start(GTK_BOX(hbox), vbox2, TRUE, TRUE, 0);
    gtk_widget_show(vbox2);

    sg = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
    sg2 = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);

    hbox2 = gtk_hbox_new(FALSE, 2);
    gtk_box_pack_start(GTK_BOX(vbox2), hbox2, FALSE, FALSE, 0);
    gtk_widget_show(hbox2);

    label = gtk_label_new_with_mnemonic("Program _name:");
    gtk_box_pack_start(GTK_BOX(hbox2), label, FALSE, FALSE, 0);
    gtk_size_group_add_widget(sg, label);
    gtk_misc_set_alignment(GTK_MISC(label), 0, 0);

    name_entry = gtk_entry_new();
    /* Set a minimum size. Since they're in a size group, the other entry will match up. */
    gtk_widget_set_size_request(name_entry, 350, -1);
    gtk_box_pack_start(GTK_BOX(hbox2), name_entry, TRUE, TRUE, 0);
    gtk_size_group_add_widget(sg2, name_entry);
    gtk_label_set_mnemonic_widget(GTK_LABEL(label), name_entry);
    gtk_widget_show(name_entry);

    hbox2 = gtk_hbox_new(FALSE, 2);
    gtk_box_pack_start(GTK_BOX(vbox2), hbox2, FALSE, FALSE, 0);
    gtk_widget_show(hbox2);
    
    label = gtk_label_new_with_mnemonic("Program _users:");
    gtk_box_pack_start(GTK_BOX(hbox2), label, FALSE, FALSE, 0);
    gtk_size_group_add_widget(sg, label);
    gtk_misc_set_alignment(GTK_MISC(label), 0, 0);

    user_entry = gtk_entry_new();
    gtk_box_pack_start(GTK_BOX(hbox2), user_entry, TRUE, TRUE, 0);
    gtk_size_group_add_widget(sg2, user_entry);
    gtk_label_set_mnemonic_widget(GTK_LABEL(label), user_entry);
    gtk_widget_show(user_entry);

    button = gtk_button_new_from_stock(GTK_STOCK_ADD);
    g_signal_connect(G_OBJECT(button), "clicked",
		     G_CALLBACK(program_add_new), NULL);
    vbox3 = gtk_vbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(hbox), vbox3, TRUE, FALSE, 0);
    gtk_widget_show(vbox3);
    gtk_box_pack_end(GTK_BOX(vbox3), button, FALSE, FALSE, 0);
    g_signal_connect(G_OBJECT(name_entry), "changed", G_CALLBACK(on_entry_changed), button);
    g_signal_connect(G_OBJECT(user_entry), "changed", G_CALLBACK(on_entry_changed), button);
    gtk_widget_set_sensitive(button, FALSE);
    gtk_widget_show(button);

    gtk_widget_show_all (ret);
    
    return ret;
    
}

static PidginPluginUiInfo ui_info = {
    get_config_frame,
    0, /* page_num (Reserved) */
    
    /* padding */
    NULL,
    NULL,
    NULL,
    NULL
};

static PurplePluginInfo info = {
    PURPLE_PLUGIN_MAGIC,
    PURPLE_MAJOR_VERSION,
    PURPLE_MINOR_VERSION,
    PURPLE_PLUGIN_STANDARD,
    PIDGIN_PLUGIN_TYPE,
    0,
    NULL,
    PURPLE_PRIORITY_DEFAULT,

    "gtk-isucan_voshtomoc",
    "voshtomoc",
    PVERSION,

    "Run external applications and use their STDIN/STDOUT in IMs.",
    "If you have a program that you interact with using STDIN/STDOUT, you can run it with this plugin and have a buddy interact with your program using an IM window. You can specify the interaction to occur with only specific buddies, for all messages or for commands only (specified with a message of the form !CMD[command]), or restrict the interaction to occur only when you are marked as away. A user can check the list of programs he/she is allowed to run using the command !LIST. Using !NEXT and !PREV, it is possible to move through the list of allowed programs.",
    
    "Ioan Alexandru Sucan",                          
    "http://ioan.sucan.info",     
    
    plugin_load,                   
    plugin_unload,
    NULL,
    &ui_info, 
    NULL, 
                         
    NULL,
    NULL,                   
    NULL,                          
    NULL,                          
    NULL,                          
    NULL                           
};                               
    
static void init_plugin(PurplePlugin *plugin)
{
}

PURPLE_INIT_PLUGIN(voshtomoc, init_plugin, info)
